/*
 * chain.c
 *
 * FAT chain I/O
 *
 * Copyright (C) 2002 Aaron Turner
 * Licensed under the GPL v2
 * See the file 'COPYING' for full licensing information
 */
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>

#include "err.h"
#include "chain.h"
#include "fatundel.h"

/*
 * this builds a FAT table chain of clusters for the file
 */
u_int16_t *
build_cluster_chain(u_int16_t start, u_int16_t len) 
{
  uint i = 0;
  u_int16_t cluster = 0;
  u_int16_t looplen;
  u_int16_t *chain;
  
  looplen = len - 2;
  chain = (u_int16_t *)calloc(sizeof(u_int16_t), (len-1)); /* the chain len is - 1 the cluster len */
  cluster = start + 1;
  
  for (i = 0; i < looplen; i ++) {
    chain[i] = cluster;
    debug(3, "chain[%04d] = 0x%04hX (%d)", i, chain[i], cluster);
    cluster ++;
  }
  chain[i] = 0xffff; /* end of file marker */
  debug(3, "chain[%04d] = 0x%hX", i, chain[i]);
  
  return chain;
}

/*
 * writes the chain to the fat at the given starting location
 */
int
write_chain(int fd, u_int16_t start, u_int16_t len, u_int16_t *chain)
{
  if (lseek(fd, start, SEEK_SET) == -1)
    err(-1, "write_chain: unable to seek", errno);

  if (write(fd, chain, (len-1)) == -1)
    err(-1, "write_chain: unable to write", errno);

  return 0;
}


/*
 * This checks the FAT table for existing files in our space
 * to see if we'll get a crosslink.
 * 0 = ok
 * > 0 = where we crosslink
 */
u_int16_t
check_crosslink(u_int16_t start, u_int16_t len, int fd)
{
  u_int16_t i = 0;
  u_int16_t data = 0;
  u_int16_t scanlen = len - 1;
  
  debug(1, "XLINK: starting scan at: %d", start);
  if (lseek(fd, start, SEEK_SET) == -1) 
    err(-1, "check_crosslink: unable to seek", errno);
  
  for (i = start; i < (start + scanlen); i ++) {
    if (read(fd, &data, sizeof(u_int16_t)) < 0)
      err(-1, "check_crosslink: readfile error", errno);
    
    debug(3, "XLINK[%d] = 0x%04hX", i, data);
    
    if (data != 0x0000) {
      warnx("check_crosslink: ran into another file at 0x%04hX", i);
      return i;
    }
    
  }
  debug(1, "XLINK: end scan");
  return 0;
}
