/*
 * fatundel.c
 *
 * FAT Undeleter for Linux
 *
 * Copyright (C) 2002 Aaron Turner
 * Licensed under the GPL v2
 * See the file 'COPYING' for full licensing information
 */

/*
 * Purpose of this program is to make it easy to undelete files 
 * requires that the 'disk' be r/o as a file 
 *
 * right now, the code only supports FAT16 and doesn't do any checking
 * to make sure you don't try this on FAT12 or FAT32 which will be 
 * pretty hosed
 */

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>

#include "err.h"
#include "fatundel.h"
#include "chain.h"
#include "file.h"

int debug_level = 0;     /* debug level */


static void usage(char *command);




int 
main(int argc, char *argv[]) 
{
  int c;
  char *filename = NULL;
  char *filesystem = NULL;
  struct stat statbuf;
  u_int16_t *chain;
  int infd = 0;
  int outfd = 0;
  int summary = 0;
  int undelete = 0;
  uint sbyte;
  //  fat_file_t file_entry;
  u_int16_t nclusters = 0; /* number of clusters */
  u_int16_t scluster = 0;  /* starting cluster */
  int override = 0;        /* override sanity checks */
  u_int16_t cluster_size = DEF_CLUSTER_SIZE;
  u_int8_t first_char = 'A';

  while ((c = getopt(argc, argv, "n:s:d:hi:f:c:SUC:")) != -1) {
    switch(c) {
    case 'C':
      cluster_size = atoi(optarg);
      break;
	case 'c':
      first_char = *optarg;
      break;
    case 'n':
      nclusters = atoi(optarg);
      break;
    case 's':
      scluster = atoi(optarg);
      break;
    case 'd':
      debug_level = atoi(optarg);
      break;
    case 'h':
      usage(argv[0]);
      break;
    case 'i':
      filesystem = optarg;
      break;
    case 'f':
      filename = optarg;
      break;
    case 'S':
      summary = 1;
      break;
    case 'U':
      undelete = 1;
      break;
    default:
      usage(argv[0]);
      break;
    }
  }

  if (nclusters < 1)
    errx(-1, "Error: -n must be > 0");
  
  if (scluster < 1)
    errx(-1, "Error: -s must be > 0");
  
  if (debug_level < 0)
    errx(-1, "Error: -d must be >= 0");
  
  if (filesystem == NULL)
    errx(-1, "Error: -f must be given");
  
  if ((infd = open(filesystem, O_RDWR)) == -1)
    err(-1, "Error: unable to open file %s", filesystem, errno);
  
  if ((fstat(infd, &statbuf)) == -1)
    err(-1, "Error: unable to stat file %s", filesystem, errno);
  
  if ((statbuf.st_size < MAX_FAT_12) && (! override))
    errx(-1, 
	 "Error: file %s is < 15MB which means it's likely a FAT12.\nUse -o to override this.", 
	 filesystem);
  
  if (((cluster_size % 512) != 0) || (cluster_size > 4096))
    errx(-1, "Error: invalid cluster size: %d", cluster_size);

  if (filename != NULL) {
     if (strcmp(filesystem, filename) == 0)
       errx(-1, "Error: -i and -f can't be the same value");
  
    if ((outfd = creat(filename, 00644)) == -1)
      err(-1, "Error: unable to open outfile %s", filename, errno);
  }

  if (undelete && summary)
    err(-1, "Error: can't use -U and -S together");

  debug(1, "cluster_size = %d", cluster_size);
  debug(1, "scluster = %d", scluster);
  debug(1, "nclusters = %d", nclusters);

  /* calc the starting byte */
  sbyte = first_cluster(infd);
  sbyte += (uint)cluster_size * ((uint)scluster - 2);
  debug(1, "sbyte = 0x%X", sbyte);


  /*
   * summary mode:
   * jump to the scluster and read it as a directory entry and
   * generate the necessary summary info for the deleted files
   * undelete mode just undels them instead of doing a report
   */
  if ((summary) || (undelete)) {
    int i, numclusters, numentries;
    void * datareturn;
    fat_file_t *dirlist;

    numentries = 0;
    if (summary)
      debug(1, "summary mode");

    if (undelete)
      debug(1, "undelete mode");

    datareturn = dir_list(sbyte, infd, cluster_size, 4);
    memcpy(&numentries, datareturn, sizeof(numentries));
    dirlist = (fat_file_t *)(datareturn + 16);
    
    debug(1, "dirlist returned %d entries", numentries);

    for (i = 0; i < numentries; i ++) {
      debug(3, "processing entry %d: %d", i, dirlist[i].name[0]);
      
      /* check for deleted file && not a dir */
      if ((dirlist[i].name[0] == 229) && (dirlist[i].fab ^ FAB_DIRECTORY)) {
        
        /* not perfect, but unless the file is exactly % 2048 it'll work */
        numclusters = dirlist[i].size / cluster_size;
        numclusters ++;
        debug(1, "[%d] Name: .%c%c%c%c%c%c%c.%c%c%c\tStart: %d\tLen: %d", 
              i, NAME2CHAR(dirlist[i].name),
              EXT2CHAR(dirlist[i].ext), dirlist[i].scluster, numclusters);

        if (summary) {
          /* build command line */
          printf("%s -i %s -s %d -n %d -f %c%c%c%c%c%c%c%c.%c%c%c\n",
                 argv[0], filesystem, dirlist[i].scluster, numclusters,
                 first_char, NAME2CHAR(dirlist[i].name), EXT2CHAR(dirlist[i].ext));
        } else if (undelete) {
          char filename[13];
          
          memset(filename, '\0', 13);
          strncpy(filename, dirlist[i].name, 8);
          strncat(filename, ".", 1);
          strncat(filename, dirlist[i].ext, 3);
          memcpy(filename, &first_char, 1);
          undelete_file(filesystem, filename, dirlist[i].scluster, 
                        numclusters, cluster_size);
        }

      } else {
        debug(2, "%s isn't deleted or is a directory", dirlist[i].name);
      }
    }

    //    free(datareturn);
  
    exit(0);
  }
  
  
  /* normal mode: */
  chain = build_cluster_chain(scluster, nclusters);
  debug(1, "Chain 5: 0x%04hX", chain[5]);
  
  sbyte = first_cluster(infd);
  
  sbyte += (uint)cluster_size * ((uint)scluster - 2); //  * sizeof(u_int16_t);
  debug(1, "sbyte = 0x%X", sbyte);
  

  /*
    if (check_crosslink(sbyte, nclusters, infd) > 0)
    exit(-1);
    
    if (write_chain(infd, sbyte, nclusters, chain) < 0)
    exit(-1);
  */

  
  if (write_file(infd, outfd, sbyte, nclusters, cluster_size) != 0)
	  errx(-1, "error writing outfile %s", filename);
  return 0;
}

void
usage(char *command) {
  fprintf(stderr, "%s -i <fs image> -n <num clusters> -s <start cluster>\n", command);
  fprintf(stderr, "\t-f <outfile> [-S|-U] -c <char> -C <cluster size> -d <debug> -h\n");
  exit(-1);
}






