/*
 * file.c
 *
 * File I/O
 *
 * Copyright (C) 2002 Aaron Turner
 * Licensed under the GPL v2
 * See the file 'COPYING' for full licensing information
 */

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>

#include "err.h"
#include "file.h"
#include "fatundel.h"

/*
 * assuming a non-fragmented file, writes a file containing
 * the data from start totaling len
 */
int 
write_file(int infd, int outfd, u_int start, u_int16_t len, u_int16_t cluster_size)
{
  int *data;
  int i;

  debug(1, "writing file");

  if (lseek(infd, start, SEEK_SET) == -1)
    err(-1, "write_file: unable to seek", errno);

  data = (int *)malloc(cluster_size);

  for (i = 0; i < len; i ++) {
    if (read(infd, data, cluster_size) == -1)
      err(-1, "write_file: unable to read", errno);

    if (write(outfd, data, cluster_size) == -1)
      err(-1, "write_file: unable to write", errno);

    debug(3, "cluster %d\t wrote %d bytes of data", i, cluster_size);

  }

  return 0;

}


/*
 * returns the byte number of the start of the root directory listing
 */
u_int
root_dir(int fd)
{
  u_int16_t secperfat;
  u_int total = 0;
  char numfat, ressec;

  if (lseek(fd, NUM_OF_FAT, SEEK_SET) == -1)
    err(-1, "get_rootdir: unable to seek", errno);
  
  if (read(fd, &numfat, 1) == -1)
    err(-1, "get_rootdir: unable to read number of FAT's", errno);

  debug(1, "number of fat = %d", numfat);

  
  if (lseek(fd, SEC_PER_FAT, SEEK_SET) == -1)
    err(-1, "get_rootdir: unable to seek", errno);
  
  if (read(fd, &secperfat, 2) == -1)
    err(-1, "get_rootdir: unable to read number of sectors per FAT", errno);

  debug(1, "sectors per fat = %hd", secperfat);

  if (lseek(fd, RESERVED_SEC, SEEK_SET) == -1)
    err(-1, "get_rootdir: unable to seek", errno);
  
  if (read(fd, &ressec, 1) == -1)
    err(-1, "get_rootdir: unable to read number of reserved sectors", errno);

  debug(1, "reserved sectors = %d", ressec);

  /* calculate root dir offset */
  total = ((u_int16_t)numfat * secperfat + (u_int16_t)ressec) * 512;

  debug(1, "root dir offsest = 0x%X", total);
  return total;

}


/*
 * returns the byte number of the start of the first cluster (cluster 2)
 */
u_int
first_cluster(int fd)
{
  u_int16_t numrootdir;
  u_int total = 0;

  total = root_dir(fd);

  if (lseek(fd, NUM_OF_ROOT_DIR, SEEK_SET) == -1)
    err(-1, "get_rootdir: unable to seek", errno);
  
  if (read(fd, &numrootdir, 2) == -1)
    err(-1, "get_rootdir: unable to read max number of root dir entries", errno);

  debug(1, "max number of entries in root = %hd", numrootdir);

  /* calc len of root dir entry and add it to the root dir offset */
  total += numrootdir * FILE_INFO_LEN;
  debug(1, "first cluster offset = 0x%X", total);

  return total;

}

/*
 * we do a really shitty thing here, we return a pointer to a void 
 * which is really a int + series of fat_file_t's
 * basically, we return the number of directory listings
 * plus a fat_file_t for each entry
 */

void *
dir_list(u_int sbyte, int fd, int cluster_size, int len)
{
  fat_file_t  *current;
  int *data;
  int entries = 0;
  int numbytes = 0;

  data = (int *)calloc(1, ((len * cluster_size) + sizeof(int)));
  current = (fat_file_t *)(data + sizeof(int));

  debug(1, "dir_list: seeking to 0x%X", sbyte);

  if (lseek(fd, sbyte, SEEK_SET) == -1)
    err(-1, "dir_list: unable to seek to start cluster: 0x%X", sbyte, errno);
  
  while (numbytes < cluster_size) {
    if (read(fd, current, sizeof(fat_file_t)) == -1)
      err(-1, "unable to read file entry", errno);
    
    numbytes += sizeof(fat_file_t);

    /* ignore null entries */
    if (current->name[0] == 0) {
      debug(2, "readdata: no file record");
      continue;
    } else {
      debug(2, "read: %c%c%c%c%c%c%c%c", current->name[0], current->name[1], 
            current->name[2], current->name[3], current->name[4],
            current->name[5], current->name[6], current->name[7]);
    }

    /* copy data over and increment counters */
    current ++;
    entries ++;
  }
  memcpy(data, &entries, sizeof(entries));
  debug(1, "return count = %d", *data);

  return (void *)data;
}

/*
 * is the equivalent of fatun -ifs -foutfile -s foo -n bar
 */

void
undelete_file(char *filesystem, char *filename, u_int16_t scluster,
                   int numclusters, int cluster_size)
{
  int infd, outfd;
  uint sbyte;

  /* open filename */
  if ((outfd = creat(filename, 00644)) == -1)
    err(-1, "Error: unable to open outfile %s", filename, errno);

  /* open filesystem */
  if ((infd = open(filesystem, O_RDWR)) == -1)
    err(-1, "Error: unable to open file %s", filesystem, errno);

  sbyte = first_cluster(infd);
  sbyte += (uint)cluster_size * ((uint)scluster -2);
  
  if (write_file(infd, outfd, sbyte, numclusters, cluster_size) != 0)
    errx(-1, "error writing outfile %s", filename);

}
